#include <iostream>
#include "scorecode.h"

//initialize the object
ScoreCode::ScoreCode(int bpm, int channels){
	m_beats_per_minute = bpm;
	m_beat_length = 60000000/bpm;
	m_channels = channels;
	m_running = false;
	if(pthread_mutex_init(&m_score_lock, NULL) != 0){
		std::cerr << "ERROR: Cannot intialize mutex" << std::endl;
		exit(1);
	}
	if(pthread_mutex_init(&m_bpm_lock, NULL) != 0){
		std::cerr << "ERROR: Cannot intialize mutex" << std::endl;
		exit(1);
	}
	
	//allocate channels
	for(int i=0; i < channels; ++i){
		m_score.push_back(std::list<char>());
		m_noteOnCb.push_back(NULL);
		m_noteOffCb.push_back(NULL);
		m_panRightCb.push_back(NULL);
		m_panLeftCb.push_back(NULL);
		m_panCenterCb.push_back(NULL);
		m_note.push_back(60);
		m_basenote.push_back(60);
		m_notestack.push_back(std::list<int>());
	}
}

//adds a channel and returns the number
int ScoreCode::addChannel(){
	m_score.push_back(std::list<char>());
	m_noteOnCb.push_back(NULL);
	m_noteOffCb.push_back(NULL);
	m_panRightCb.push_back(NULL);
	m_panLeftCb.push_back(NULL);
	m_panCenterCb.push_back(NULL);
	m_note.push_back(60);
	m_basenote.push_back(60);
	m_notestack.push_back(std::list<int>());
	return m_channels++;
}

//removes a channel and returns number of channels
int ScoreCode::removeChannel(){
	if(m_channels > 0){
		m_score.pop_back();
		m_noteOnCb.pop_back();
		m_noteOffCb.pop_back();
		m_panRightCb.pop_back();
		m_panLeftCb.pop_back();
		m_panCenterCb.pop_back();
		m_note.pop_back();
		m_basenote.pop_back();
		m_notestack.pop_back();
		return --m_channels;
	}else{
		return 0;
	}
}

//clears all scorecode on a channel
void ScoreCode::clearChannel(int channel){
	if(pthread_mutex_lock(&m_score_lock) != 0){
		std::cerr << "ERROR: Could not acquire mutex lock" << std::endl;
		exit(1);
	}
	m_score[channel].clear();
	if(pthread_mutex_unlock(&m_score_lock) != 0){
		std::cerr << "ERROR: Could not unlock mutex" << std::endl;
		exit(1);
	}
}

//clears all channels of all scorecode
void ScoreCode::clearAllChannels(){
	if(pthread_mutex_lock(&m_score_lock) != 0){
		std::cerr << "ERROR: Could not acquire mutex lock" << std::endl;
		exit(1);
	}
	for(int i=0; i < m_channels; i++){
		m_score[i].clear();
	}
	if(pthread_mutex_unlock(&m_score_lock) != 0){
		std::cerr << "ERROR: Could not unlock mutex" << std::endl;
		exit(1);
	}
}

//returns the number of channels
int ScoreCode::getNumChannels(){
	return m_channels;
}

//push a pattern onto a channel
void ScoreCode::pushPattern(int channel, std::string p){
	//lock the score
	if(pthread_mutex_lock(&m_score_lock) != 0){
		std::cerr << "ERROR: Could not acquire mutex lock" << std::endl;
		exit(1);
	}
	
	//push a pattern one note at a time
	for(int i=0; i < p.length(); ++i){
		m_score[channel].push_back(p[i]);
	}
	
	//unlock score
	if(pthread_mutex_unlock(&m_score_lock) != 0){
		std::cerr << "ERROR: Could not unlock mutex" << std::endl;
		exit(1);
	}
}

//sets the BPM
void ScoreCode::setBPM(int bpm){
	if(pthread_mutex_lock(&m_bpm_lock) != 0){
		std::cerr << "ERROR: Could not acquire mutex lock" << std::endl;
		exit(1);
	}
	m_beats_per_minute = bpm;
	m_beat_length = 60000000/bpm;
	if(pthread_mutex_unlock(&m_bpm_lock) != 0){
		std::cerr << "ERROR: Could not unlock mutex" << std::endl;
		exit(1);
	}
}

//gets BPM
int ScoreCode::getBPM(){
	return m_beats_per_minute;
}

//sets the base note for a channel
void ScoreCode::setBase(int channel, int note){
	m_basenote[channel] = note;
}

//gets the base note for a channel
int ScoreCode::getBase(int channel){
	return m_basenote[channel];
}

//note callback registrations
void ScoreCode::registerNoteOnCallback(int channel, noteOnCallback cb){
	m_noteOnCb[channel] = cb;
}

void ScoreCode::registerNoteOffCallback(int channel, noteOffCallback cb){
	m_noteOffCb[channel] = cb;
}

void ScoreCode::registerPanRightCallback(int channel, panRightCallback cb){
	m_panRightCb[channel] = cb;
}

void ScoreCode::registerPanLeftCallback(int channel, panLeftCallback cb){
	m_panLeftCb[channel] = cb;
}

void ScoreCode::registerPanCenterCallback(int channel, panCenterCallback cb){
	m_panCenterCb[channel] = cb;
}

//returns the current note, taking the stack into account
int ScoreCode::getNote(int channel){
	if(m_notestack[channel].size() > 0)
		return m_notestack[channel].back();
	else
		return m_note[channel];
}

//sets the current note to the passed value, taking stack into account
void ScoreCode::setNote(int channel, int note){
	if(m_notestack[channel].size() > 0)
		m_notestack[channel].back() = note;
	else
		m_note[channel] = note;
}


//the method that gets wrapped and runs the interpreter
void *ScoreCode::startInt(void *arg){
	ScoreCode *s = (ScoreCode*)arg;
	s->interpret();
	return 0;
}

//the run function, actually write this...
void ScoreCode::run(){
	if(!m_running){
		m_running = true;
		if(pthread_create(&m_intThread, NULL, &startInt, this) != 0){
			std::cerr << "ERROR: Could not create thread" << std::endl;
			exit(1);
		}
	}
}

//the stop function, actually write this...
void ScoreCode::stop(){
	m_running = false;
	pthread_join(m_intThread, NULL);
}

//returns whether or not the interpreter is running
bool ScoreCode::isRunning(){
	return m_running;
}

//function that actually does stuff, the interpreter
void ScoreCode::interpret(){
	//run until stopped
	while(m_running){
		//iterate through all channels and process their events until we reach a rest
		for(int i=0; i < m_channels; i++){
			bool done = false;
			while(!done && m_score[i].size() > 0){
				//lock the score
				if(pthread_mutex_lock(&m_score_lock) != 0){
					std::cerr << "ERROR: Could not acquire mutex lock" << std::endl;
					exit(1);
				}
				
				char c = m_score[i].front();
				m_score[i].pop_front();
				switch(c){
					case 'o': //note on
						if(m_noteOnCb[i] != NULL)
							(*m_noteOnCb[i])(getNote(1));
						break;
					case 'x': //note off
						if(m_noteOffCb[i] != NULL)
							(*m_noteOffCb[i])(getNote(i));
						break;
					case '+': //pitch up
						setNote(i, getNote(i) + 1);
						break;
					case '-': //pitch down
						setNote(i, getNote(i) - 1);
						break;
					case '!': //reset pitch
						setNote(i, m_basenote[i]);
						break;
					case '.': //rest
						done = true;
						break;
					case '/': //pan right
						if(m_panRightCb[i] != NULL)
							(*m_panRightCb[i])();
						break;
					case '\\': //pan left
						if(m_panLeftCb[i] != NULL)
							(*m_panLeftCb[i])();
						break;
					case '|': //pan center
						if(m_panCenterCb[i] != NULL)
							(*m_panCenterCb[i])();
						break;
					case '[':
						m_notestack[i].push_back(getNote(i));
						break;
					case ']':
						if(m_notestack[i].size() > 0)
							m_notestack[i].pop_back();
						break;
					case '%':
						m_notestack[i].clear();
						setNote(i, m_basenote[i]);
						break;
					default:
						break;
				}
				//unlock the score
				if(pthread_mutex_unlock(&m_score_lock) != 0){
					std::cerr << "ERROR: Could not unlock mutex" << std::endl;
					exit(1);
				}
			}
		}
		
		//delay here for the rest
		if(pthread_mutex_lock(&m_bpm_lock) != 0){
			std::cerr << "ERROR: Could not acquire mutex lock" << std::endl;
			exit(1);
		}
		usleep(m_beat_length);
		if(pthread_mutex_unlock(&m_bpm_lock) != 0){
			std::cerr << "ERROR: Could not unlock mutex" << std::endl;
			exit(1);
		}
	}
}
